/*
*	TITLE		Queued serial drivers for FreeRTOS using PIC32MZ
*
*   File        serial_RTOS.c
*
*	VERSION:	0.3b (Beta)
*
*	DATE:		3-Dec-2013
*
*	AUTHOR:		Tom Lafleur  
*				lafleur@lafleur.us
*
*	COMMENTS:	You are free to use this source code for any non-commercial
*				use. Please do not make copies of this source code, modified
*				or un-modified, publicly available on the internet or
*				elsewhere without permission. Thanks.
*
*			Copyright 2005-2014  All rights are reserved.
*
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	12-Jun-2009  0.1  TRL - Original code for N2K60 under FreeRTOS
*	01-Oct-2011  0.2a TRL - Added serial port 2, write function, error counters
*	15-Mar-2012  0.2b TRL - Clean up issues with .h files
*	03-Dec-2013	 0.3b TRL - Port to PIC32MZ processor
*
*
*	Notes:	1)  Tested with FreeRTOS ver 7.6.0 MEB II
*			2)  Support for four serial ports, can easily be expanded for more ports
*			3)  Added Rx error counters to the data structures
*			4)  If using processors with Peripheral pin select, you must define
*                 pin's prior to using this driver
*			5)  Based on FreeRTOS demo code and my 2007 robotics serial driver  
*
*	ToDo:	1)  Check if code is atomic and thread safe
*			2)  Make Rx error counters avilable global
*			3)  Make Tx use DMA if avilable, if so we will use ring buffers
*               and give a full buffer to DMA system. Queue's are too slow.
*			4) 
*
* 	My Notes:
*		
*		
*		
*/

/* ***** Example of use ***** */
/*
    const unsigned portBASE_TYPE uxTxBufferLength = 32;	// Tx queues size
	const unsigned portBASE_TYPE uxRxBufferLength = 64;	// Rx queues size

	// Initialise the com port then spawn the Rx and Tx tasks.

	xPort = xSerialPortInit ( serCOM4, eBaudRate, serNO_PARITY, serBITS_8, serSTOP_1, uxTxBufferLength, uxRxBufferLength );

	// The Tx task is spawned with a lower priority than the Rx task.
	xTaskCreate( vComTxTask, ( signed char * ) "COMTx4", comSTACK_SIZE, NULL, uxPriority - 1, ( xTaskHandle * ) NULL );
	xTaskCreate( vComRxTask, ( signed char * ) "COMRx4", comSTACK_SIZE, NULL, uxPriority, ( xTaskHandle * ) NULL );
*/


/* ***** If using processors with PPS, Peripheral pin select, you must define pin's prior to using this driver *****  */
/* ***** For now we will define PPS bits here, but should be defined in bsp.c file ***** */


/* FreeRTOS Scheduler include files. */
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"

/* System include files */
#include "hardware.h"

/* Serial include files. */
#include "serial.h"
#include "serial_RTOS.h"

#include "diagprint.h"

#if defined (__PIC32MZ)
//#warning "Using PIC32MZ serial driver!"
#else
#warning "This driver is for PIC32MZ and it is not defined!!!!!!"
#endif

/*-----------------------------------------------------------*/
/*  static configuration of serial port from these defines */

#define serNO_BLOCK			( ( portTickType ) 0 )
#define COMQUEUE			// add queues to registry for debug

#define serMAX_PORTS		((unsigned short) 4 )	// max number of serial ports we are supporting

#define hwUNLOCK_KEY_0		( 0xAA996655UL )
#define hwUNLOCK_KEY_1		( 0x556699AAUL )

/*-----------------------------------------------------------*/
/*     Prototype for private functions defined here */
void prvSerialPort1Init(unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits);
void prvSerialPort2Init(unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits);
void prvSerialPort3Init(unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits);
void prvSerialPort4Init(unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits);

/*-----------------------------------------------------------*/
/*  Port structure */
static xComPort xPorts[serMAX_PORTS] =		 // up to three serial ports at this time. zero is UART1. ect...
{        

    {0, NULL, NULL, 0, 0, 0, 0},    // Serial Port 1, xRxedChars handle, xCharsForTx handle, xTxHasEnded, Party, Framing, Overrun
    {1, NULL, NULL, 0, 0, 0, 0},    // Serial Port 2, xRxedChars handle, xCharsForTx handle, xTxHasEnded, Party, Framing, Overrun
    {2, NULL, NULL, 0, 0, 0, 0},  	// Serial Port 3, xRxedChars handle, xCharsForTx handle, xTxHasEnded, Party, Framing, Overrun
    {3, NULL, NULL, 0, 0, 0, 0}		// Serial Port 4, xRxedChars handle, xCharsForTx handle, xTxHasEnded, Party, Framing, Overrun
};

/*-----------------------------------------------------------*/

/* The UART interrupt handler.  As this uses the FreeRTOS assembly interrupt
entry point the IPL setting in the following prototypes has no effect.  The
interrupt priority is set by the call to  ConfigIntUARTx () in
xSerialPortInitMinimal(). */

#if defined UART1
void __attribute__( (interrupt(ipl0), vector(_UART1_RX_VECTOR)))	vU1RXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART1_TX_VECTOR)))	vU1TXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART1_FAULT_VECTOR))) vU1FAULTInterruptWrapper( void );
#endif

#if defined UART2
void __attribute__( (interrupt(ipl0), vector(_UART2_RX_VECTOR)))	vU2RXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART2_TX_VECTOR)))	vU2TXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART2_FAULT_VECTOR))) vU2FAULTInterruptWrapper( void );
#endif

#if defined UART3
void __attribute__( (interrupt(ipl0), vector(_UART3_RX_VECTOR)))	vU3RXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART3_TX_VECTOR)))	vU3TXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART3_FAULT_VECTOR))) vU3FAULTInterruptWrapper( void );
#endif

#if defined UART4
void __attribute__( (interrupt(ipl0), vector(_UART4_RX_VECTOR)))	vU4RXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART4_TX_VECTOR)))	vU4TXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART4_FAULT_VECTOR))) vU4FAULTInterruptWrapper( void );
#endif


/*-----------------------------------------------------------*/
xComPortHandle xSerialPortInit(eCOMPort ePort, unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits, unsigned portBASE_TYPE uxTxBufferLength, unsigned portBASE_TYPE uxRxBufferLength)
{
    portENTER_CRITICAL(); // for now we will make sure we have control
    {
        if (ePort < serMAX_PORTS) 
        	{
            switch (ePort) 
                {
#if defined UART1
                case 0:
                    prvSerialPort1Init(eWantedBaud, eWantedParity, eWantedDataBits, eWantedStopBits);
                    break;
#endif
#if defined UART2
                case 1:
                    prvSerialPort2Init(eWantedBaud, eWantedParity, eWantedDataBits, eWantedStopBits);
                    break;
#endif
#if defined UART3
                case 2:
                    prvSerialPort3Init(eWantedBaud, eWantedParity, eWantedDataBits, eWantedStopBits);
                    break;
#endif
#if defined UART4
                case 3:
                    prvSerialPort4Init(eWantedBaud, eWantedParity, eWantedDataBits, eWantedStopBits);
                    break;
#endif
                default:
                    break;
                }

			/* Create the serial queues used by the tasks. */ // returns NULL is unsuccessfully
            xPorts[ePort].xRxedChars  = xQueueCreate(uxRxBufferLength, (unsigned portBASE_TYPE) sizeof (signed portCHAR));
				if (xPorts[ePort].xRxedChars == NULL)
					{
						// if error on create
					}
            xPorts[ePort].xCharsForTx = xQueueCreate(uxTxBufferLength, (unsigned portBASE_TYPE) sizeof (signed portCHAR));
				if (xPorts[ePort].xCharsForTx == NULL)
					{
						// if error on create
					}

#if defined (COMQUEUE)		// used's only for debug
            switch (ePort) 
                {
  #if defined UART1
                case 0:
                    vQueueAddToRegistry(xPorts[ePort].xRxedChars,  "Rx1 Queue");
                    vQueueAddToRegistry(xPorts[ePort].xCharsForTx, "Tx1 Queue");
                    break;
  #endif
  #if defined UART2
                case 1:
                    vQueueAddToRegistry(xPorts[ePort].xRxedChars,  "Rx2 Queue");
                    vQueueAddToRegistry(xPorts[ePort].xCharsForTx, "Tx2 Queue");
                    break;
  #endif
  #if defined UART3
               	case 2:
                    vQueueAddToRegistry(xPorts[ePort].xRxedChars,  "Rx3 Queue");
                    vQueueAddToRegistry(xPorts[ePort].xCharsForTx, "Tx3 Queue");
                    break;
  #endif
  #if defined UART4
               	case 3:
                    vQueueAddToRegistry(xPorts[ePort].xRxedChars,  "Rx4 Queue");
                    vQueueAddToRegistry(xPorts[ePort].xCharsForTx, "Tx4 Queue");
                    break;
  #endif
                default:
                    break;
                }
#endif  // (COMQUEUE)

        }
    }
    portEXIT_CRITICAL();
    return &xPorts[ePort];
}

/* ********************* UART 1 Initialization ********************* */
#if defined UART1

void prvSerialPort1Init( unsigned long ulWantedBaud, eParity Parity, eDataBits DataBits, eStopBits StopBits)
{

/* It is assumed that this function is called prior to the scheduler being
    started.  Therefore interrupts must not be allowed to occur yet as they
    may attempt to perform a context switch. */

    portDISABLE_INTERRUPTS();

	char cChar;

/* For now we will define PPS bits here, but should be defined in bsp.c file */
// Note: AD1PCFG is now --> ANSELx, x = Port bank, 1=analog, 0-digital, Default is 0xffff
	ANSELGbits.ANSG8 = 0;

	SYSKEY = 0;					// make sure its locked
	SYSKEY = hwUNLOCK_KEY_0;	// now unlock it
	SYSKEY = hwUNLOCK_KEY_1;

#if defined MEB2
	U1RXR = 1;		// RPG8 --> U1RX  Pin 9  on J2 of MEB II board
	RPB3R = 1;		// U1TX --> RPB3  Pin 10 on J2 of MEB II board
#elif WMR2MZ

#endif
	SYSKEY = 0;		// lets relock it


	/* Configure the UART and its interrupts. */
	U1BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)16 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 0
//  U1BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)4 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 1

	U1MODEbits.BRGH = 0;		// 0 = 16X clock, 1 = 4X clock

	if ( DataBits == serBITS_9) DataBits = 0x03;	// 9-bit data, no parity
	U1MODEbits.PDSEL = ((Parity || DataBits) && 0x03);

	U1MODEbits.STSEL = StopBits;    //Stop Bit Selection bit


	U1STAbits.URXISEL = 0b00;	// RX Interrupt when buffer has a char (0b10 = 3/4 full)
	U1STAbits.UTXISEL = 0b00;	// TX Interrupt when buffer empty,

	U1STAbits.OERR = 0;			// Clear any errors
	U1STAbits.FERR = 0;
	U1STAbits.PERR = 0;

	while (U1STAbits.URXDA == 1)  cChar = U1RXREG;	// Clear FIFO

	U1MODEbits.LPBACK = 0;  	// Loopback bit

	IEC3bits.U1RXIE = 1;		// Enable interrupts
	IEC3bits.U1EIE =  1;		// Error interrupt
    /* Turn off UART TX interrupt, we will turn it on in the Queue handler */
    IEC3bits.U1TXIE = 0;

	IPC28bits.U1RXIP = configKERNEL_INTERRUPT_PRIORITY + 1;	// priority 0-->7
	IPC28bits.U1TXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC28bits.U1EIP =  configKERNEL_INTERRUPT_PRIORITY + 1;

	IPC28bits.U1RXIS = 1;		// sub priority 0-->3
	IPC28bits.U1TXIS = 0;
	IPC28bits.U1EIS =  0;

	U1STAbits.URXEN = 1;		// Rx Enable
	U1STAbits.UTXEN = 1;		// Tx Enable

	IFS3bits.U1RXIF = 0;		// Clear any interrupts that might be pending
	IFS3bits.U1TXIF = 0;
	IFS3bits.U1EIF =  0;

	U1MODEbits.ON =   1;		// UART enable

    xPorts[0].xTxHasEnded = pdTRUE;
}

#endif

/* ********************* UART 2 Initialization ********************* */

#if defined UART2

/*-----------------------------------------------------------*/
void prvSerialPort2Init( unsigned long ulWantedBaud, eParity Parity, eDataBits DataBits, eStopBits StopBits)
{

/* It is assumed that this function is called prior to the scheduler being
    started.  Therefore interrupts must not be allowed to occur yet as they
    may attempt to perform a context switch. */

    portDISABLE_INTERRUPTS();

	char cChar;

/* For now we will define PPS bits here, but should be defined in bsp.c file */
// Note: AD1PCFG is now --> ANSELx, x = Port bank, 1=analog, 0-digital, Default is 0xffff
	ANSELGbits.ANSG6 = 0;

	SYSKEY = 0;					// make sure its locked
	SYSKEY = hwUNLOCK_KEY_0;	// now unlock it
	SYSKEY = hwUNLOCK_KEY_1;

#if defined MEB2
	U2RXR = 1;		// RPG6 --> U2RX  Pin 11 on J2 of MEB II board
	RPB14R = 2;		// U2TX --> RPB14 Pin 19 on J2 of MEB II board
#elif WMR2MZ

#endif
	SYSKEY = 0;		// lets relock it

	/* Configure the UART and its interrupts. */
	U2BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)16 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 0
	// U2BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)4 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 1

	U2MODEbits.BRGH = 0;		// 0 = 16X clock, 1 = 4X clock
		
	if ( DataBits == serBITS_9) DataBits = 0x03;	// 9-bit data, no parity
	U2MODEbits.PDSEL = ((Parity || DataBits) && 0x03);

	U2MODEbits.STSEL = StopBits;    //Stop Bit Selection bit

	
	U2STAbits.URXISEL = 0b00;	// RX Interrupt when buffer has a char (0b10 = 3/4 full)
	U2STAbits.UTXISEL = 0b00;	// TX Interrupt when buffer empty,

	U2STAbits.OERR = 0;			// Clear any errors
	U2STAbits.FERR = 0;
	U2STAbits.PERR = 0;
	
	while (U2STAbits.URXDA == 1)  cChar = U2RXREG;	// Clear FIFO

	U2MODEbits.LPBACK = 0;  	// Loopback bit

	IEC4bits.U2RXIE = 1;		// Enable interrupts
	IEC4bits.U2EIE =  1;		// Error interrupt
    /* Turn off UART TX interrupt, we will turn it on in the Queue handler */
    IEC4bits.U2TXIE = 0;

	IPC36bits.U2RXIP = configKERNEL_INTERRUPT_PRIORITY + 1;	// priority 0-->7
	IPC36bits.U2TXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC36bits.U2EIP =  configKERNEL_INTERRUPT_PRIORITY + 1;

	IPC36bits.U2RXIS = 1;		// sub priority 0-->3
	IPC36bits.U2TXIS = 0;
	IPC36bits.U2EIS =  0;

	U2STAbits.URXEN = 1;		// Rx Enable
	U2STAbits.UTXEN = 1;		// Tx Enable
	
	IFS4bits.U2RXIF = 0;		// Clear any interrupts that might be pending
	IFS4bits.U2TXIF = 0;
	IFS4bits.U2EIF =  0;

	U2MODEbits.ON =   1;		// UART enable

    xPorts[1].xTxHasEnded = pdTRUE;
}
#endif

/* ********************* UART 3 Initialization ********************* */

#if defined UART3

void prvSerialPort3Init( unsigned long ulWantedBaud, eParity Parity, eDataBits DataBits, eStopBits StopBits)
{

/* It is assumed that this function is called prior to the scheduler being
    started.  Therefore interrupts must not be allowed to occur yet as they
    may attempt to perform a context switch. */

    portDISABLE_INTERRUPTS();

	char cChar;

/* For now we will define PPS bits here, but should be defined in bsp.c file */
// Note: AD1PCFG is now --> ANSELx, x = Port bank, 1=analog, 0-digital, Default is 0xffff
	ANSELGbits.ANSG7 = 0;

	SYSKEY = 0;					// make sure its locked
	SYSKEY = hwUNLOCK_KEY_0;	// now unlock it
	SYSKEY = hwUNLOCK_KEY_1;

#if defined MEB2
	U3RXR = 1;		// RPG7 --> U3RX  Pin 9  on J2 of MEB II board
	RPB10R = 1;		// U3TX --> RPB10 Pin 21 on J2 of MEB II board
#elif WMR2MZ

#endif
	SYSKEY = 0;		// lets relock it

	/* Configure the UART and its interrupts. */
	U3BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)16 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 0
//  U3BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)4 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 1

	U3MODEbits.BRGH = 0;		// 0 = 16X clock, 1 = 4X clock

	if ( DataBits == serBITS_9) DataBits = 0x03;	// 9-bit data, no parity
	U3MODEbits.PDSEL = ((Parity || DataBits) && 0x03);

	U3MODEbits.STSEL = StopBits;    //Stop Bit Selection bit


	U3STAbits.URXISEL = 0b00;	// RX Interrupt when buffer has a char (0b10 = 3/4 full)
	U3STAbits.UTXISEL = 0b00;	// TX Interrupt when buffer empty,

	U3STAbits.OERR = 0;			// Clear any errors
	U3STAbits.FERR = 0;
	U3STAbits.PERR = 0;

	while (U3STAbits.URXDA == 1)  cChar = U3RXREG;	// Clear FIFO

	U3MODEbits.LPBACK = 0;  	// Loopback bit

	IEC4bits.U3RXIE = 1;		// Enable interrupts
	IEC4bits.U3EIE =  1;		// Error interrupt
    /* Turn off UART TX interrupt, we will turn it on in the Queue handler */
    IEC4bits.U3TXIE = 0;

	IPC39bits.U3RXIP = configKERNEL_INTERRUPT_PRIORITY + 1;	// priority 0-->7
	IPC39bits.U3TXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC39bits.U3EIP =  configKERNEL_INTERRUPT_PRIORITY + 1;

	IPC39bits.U3RXIS = 1;		// sub priority 0-->3
	IPC39bits.U3TXIS = 0;
	IPC39bits.U3EIS =  0;

	U3STAbits.URXEN = 1;		// Rx Enable
	U3STAbits.UTXEN = 1;		// Tx Enable

	IFS4bits.U3RXIF = 0;		// Clear any interrupts that might be pending
	IFS4bits.U3TXIF = 0;
	IFS4bits.U3EIF =  0;

	U3MODEbits.ON =   1;		// UART enable

    xPorts[2].xTxHasEnded = pdTRUE;
}

#endif


/* ********************* UART 4 Initialization ********************* */
#if defined UART4

void prvSerialPort4Init( unsigned long ulWantedBaud, eParity Parity, eDataBits DataBits, eStopBits StopBits)
{

/* It is assumed that this function is called prior to the scheduler being
    started.  Therefore interrupts must not be allowed to occur yet as they
    may attempt to perform a context switch. */

    portDISABLE_INTERRUPTS();

	char cChar;

/* For now we will define PPS bits here, but should be defined in bsp.c file */
// Note: AD1PCFG is now --> ANSELx, x = Port bank, 1=analog, 0-digital, Default is 0xffff
//	ANSELDbits.ANSD1 = 0;

	SYSKEY = 0;					// make sure its locked
	SYSKEY = hwUNLOCK_KEY_0;	// now unlock it
	SYSKEY = hwUNLOCK_KEY_1;

#if defined MEB2
	U4RXR = 0;		// RPD1 --> U4RX  Pin 25 on J2 of MEB II board
	RPB0R = 2;		// U4TX --> RPB0  Pin 23 on J2 of MEB II board
#elif WMR2MZ

#endif
	SYSKEY = 0;		// lets relock it

	/* Configure the UART and its interrupts. */
	U4BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)16 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 0
//  U4BRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)4 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 1

	U4MODEbits.BRGH = 0;		// 0 = 16X clock, 1 = 4X clock

	if ( DataBits == serBITS_9) DataBits = 0x03;	// 9-bit data, no parity
	U4MODEbits.PDSEL = ((Parity || DataBits) && 0x03);

	U4MODEbits.STSEL = StopBits;    //Stop Bit Selection bit


	U4STAbits.URXISEL = 0b00;	// RX Interrupt when buffer has a char (0b10 = 3/4 full)
	U4STAbits.UTXISEL = 0b00;	// TX Interrupt when buffer empty,

	U4STAbits.OERR = 0;			// Clear any errors
	U4STAbits.FERR = 0;
	U4STAbits.PERR = 0;

	while (U4STAbits.URXDA == 1)  cChar = U4RXREG;	// Clear FIFO

	U4MODEbits.LPBACK = 0;  	// Loopback bit

	IEC5bits.U4RXIE = 1;		// Enable interrupts
	IEC5bits.U4EIE =  1;		// Error interrupt
    /* Turn off UART TX interrupt, we will turn it on in the Queue handler */
    IEC5bits.U4TXIE = 0;

	IPC42bits.U4RXIP = configKERNEL_INTERRUPT_PRIORITY + 1;	// priority 0-->7
	IPC43bits.U4TXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC42bits.U4EIP =  configKERNEL_INTERRUPT_PRIORITY + 1;

	IPC42bits.U4RXIS = 1;		// sub priority 0-->3
	IPC43bits.U4TXIS = 0;
	IPC42bits.U4EIS =  0;

	U4STAbits.URXEN = 1;		// Rx Enable
	U4STAbits.UTXEN = 1;		// Tx Enable

	IFS5bits.U4RXIF = 0;		// Clear any interrupts that might be pending
	IFS5bits.U4TXIF = 0;
	IFS5bits.U4EIF =  0;

	U4MODEbits.ON =   1;		// UART enable

    xPorts[3].xTxHasEnded = pdTRUE;
}

#endif

/*-----------------------------------------------------------*/
signed portBASE_TYPE xSerialGetChar(xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime)
{
    /* Get the next character from the buffer.  Return false if no characters
    are available or arrive before xBlockTime expires. */

	if (xQueueReceive(((xComPort *) pxPort)->xRxedChars, pcRxedChar, xBlockTime))
	{
        return pdTRUE;
    } 
	else
	{
        return pdFALSE;
    }
}

/* define some standard ASCII char */
#define BS_char		0x08
#define LF_char		0x0a
#define CR_char		0x0d
#define DEL_char	0x7f
#define SPACE_char	0x20
#define ESC_char	0x1b

/* This is a special version for the FreeRTOS-CLI processor that echo's all char's, strips CR, LF and handles BS */
/*-----------------------------------------------------------*/
void xSerialGetStringS(xComPortHandle pxPort, signed char * pcString, unsigned short usStringLength, portTickType xBlockTime)
{
	char c;
	int char_count = 0;

    /* Call to get a NULL or CR ( termintated string from the queue of max length usStringLength
	 * Note: on a PC and some VT100 terminals, DELETE key sends an "esc[3~", this is NOT the same as a BS */
	do
	{
		xQueueReceive(((xComPort *) pxPort)->xRxedChars, &c, xBlockTime);
			
		if ( !(c == LF_char) )		// ignore LF from input
		{
			char_count++;			// we just got a char, sot increment char count
			if ( ( c == BS_char) || (c == DEL_char) )
			{
				if (char_count > 1)	// need to make sure we donot backup pass begining of pcString buffer
				{
					char_count--;		// adjust for BS char
					char_count--;		// adjust for last char
					pcString--;			// adjust for BS in buffer
					usStringLength++;	// adjust length
					xSerialPutChar(pxPort, BS_char , serNO_BLOCK);		// we need to send a back space
					xSerialPutChar(pxPort, SPACE_char , serNO_BLOCK);	// we need to clear char
					if (!(c == BS_char))	xSerialPutChar(pxPort, BS_char , serNO_BLOCK);		// we need to send another back space
				}
				else
				{
					char_count = 0;
				}
			}

			else
			{
				*pcString = c;			// send char to pcString buffer
				if (c == CR_char) *pcString = NULL;		// CR --> NULL
				pcString++;				// adjust pointer for next char
				usStringLength--;		// adjust max length
			}

			xSerialPutChar(pxPort, c, serNO_BLOCK);		// echo char back to serial port
		}
	
	}	while (( c != NULL) && (c != 0x0d) && (usStringLength > 0));

	return;
}

/*-----------------------------------------------------------*/
void xSerialGetString(xComPortHandle pxPort, signed char * pcString, unsigned short usStringLength, portTickType xBlockTime)
{
	char c;

    /* Call to get a NULL or CR termintated string from the queue of max length usStringLength */
	do
	{
		xQueueReceive(((xComPort *) pxPort)->xRxedChars, &c, xBlockTime);
		if ( !(c == LF_char) )		// ignore LF from input
		{
		usStringLength--;
		*pcString = c;
		if (c == CR_char) *pcString = NULL;	// convert CR to NULL
		pcString++;
		}

	}	while (( c != NULL) && (c != CR_char) && (usStringLength > 0)); 

	return;
}

/*-----------------------------------------------------------*/
signed portBASE_TYPE xSerialPeekChar(xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime)
{
    /* Get the next character from the buffer.  Return false if no characters
    are available or arrive before xBlockTime expires. */

	if (xQueuePeek(((xComPort *) pxPort)->xRxedChars, pcRxedChar, xBlockTime))
	{
        return pdTRUE;
    }
	else
	{
        return pdFALSE;
    }
}


/*-----------------------------------------------------------*/
signed portBASE_TYPE xSerialPutChar(xComPortHandle pxPort, signed char cOutChar, portTickType xBlockTime)
{
    /* Return false if after the block time there is no room on the Tx queue. */
	
    if (xQueueSend(((xComPort *)pxPort)->xCharsForTx, &cOutChar, xBlockTime) != pdPASS)
	{
        return pdFAIL;
    }
    /* A critical section should not be required as xTxHasEnded will not be
    written to by the ISR if it is already 0 (is this correct?) */

    if (pxPort->xTxHasEnded)
	{
        pxPort->xTxHasEnded = pdFALSE;
        
        switch (pxPort->sPort)
                {
#if defined UART1
                	case 0:
					IEC3SET = _IEC3_U1TXIE_MASK;
                	break;
#endif
#if defined UART2                	
                case 1:
					IEC4SET = _IEC4_U2TXIE_MASK;
   					break;
#endif
#if defined UART3                
                case 2:
					IEC4SET = _IEC4_U3TXIE_MASK;
   					break;
#endif
#if defined UART4
                case 3:
					IEC5SET = _IEC5_U4TXIE_MASK;
   					break;
#endif
   				default:
                    break;
    	}
    }
    return pdPASS;
}

// do we need to add block time here ?? <----------------------
/*-----------------------------------------------------------*/
void vSerialPutString(xComPortHandle pxPort, const signed char * const pcString, unsigned short usStringLength)
{
    signed char *pxNext;

    /* A parameters that this port does not use. */
    (void) usStringLength;

    /* Send each character in the string, one at a time. */
    pxNext = (signed char *) pcString;
    while (*pxNext)
	{
        xSerialPutChar(pxPort, *pxNext, serNO_BLOCK);
        pxNext++;
    }
}


/*-----------------------------------------------------------*/
void vSerialPutStringN(xComPortHandle pxPort, const signed char * const pcString, unsigned short usStringLength)
{
    signed char *pxNext;

    /* Send each character in the string, one at a time until usStringLength = 0. */
    pxNext = (signed char *) pcString;
    while ( (*pxNext) && (usStringLength >0 ) )
	{
        xSerialPutChar(pxPort, *pxNext, serNO_BLOCK);
        pxNext++;
		usStringLength--;
    }
}

/**********************************************************************
 *
 *	This allows the used of the generic C STDIO features to write to another port:
 *	FILE *port = fopen("1","w");
 *	fprintf(port,"This will be output through the serial port \n");
 *
 *	Also use by printf with a handle of 1
 *
 *	Added support for basic serial port 1 and 2 using FreeRTOS Queue's
 *	case 3 will print to serial port 2, if it open
 *	This replaces the standard write.c routine in the XC30/XC32 lib
 *  Warning, check to make sure your C STDIO are RTOS safe!!
 *
 **********************************************************************/

int write(int handle, void *buffer, unsigned int len)
{
    int i;
    switch (handle) {
        default:
        case 0: // stdin
        case 1: // stdout
        case 2: // stderr

#if defined UART1	// serial port 1
            // send the data to serial port 1 via FreeRTOS
            if (!(xPorts[0].xCharsForTx == NULL)) {
                for (i = len; i; --i) {
                    xSerialPutChar(&(xPorts[0]), *(char*) buffer, 0);
                    break;
                }
            }
            break;
#endif
#if defined UART2
        case 3: // serial port 2

            // send the data to serial port 2 via FreeRTOS
            if (!(xPorts[1].xCharsForTx == NULL)) {
                for (i = len; i; --i) {
                    xSerialPutChar(&(xPorts[1]), *(char*) buffer, 0);
                    break;
                }
            }
            break;
#endif
#if defined UART3
        case 4: // serial port 3

            // send the data to serial port 3 via FreeRTOS
            if (!(xPorts[2].xCharsForTx == NULL)) {
                for (i = len; i; --i) {
                    xSerialPutChar(&(xPorts[2]), *(char*) buffer, 0);
                    break;
                }
            }
            break;
#endif
#if defined UART4
			case 5: // serial port 4

            // send the data to serial port 3 via FreeRTOS
            if (!(xPorts[3].xCharsForTx == NULL)) {
                for (i = len; i; --i) {
                    xSerialPutChar(&(xPorts[3]), *(char*) buffer, 0);
                    break;
                }
            }
            break;
#endif
    }
    return (len);
}

/*-----------------------------------------------------------*/

/** 
 * Close serial port. Turn off the interrupts.
 */

void vSerialClose(xComPortHandle xPort)
{
    unsigned short usPort;

    usPort = ((xComPort *) xPort)->sPort;

    switch (usPort)
	{
        case 0:
			IEC3CLR = _IEC3_U1TXIE_MASK;	// Disable interrupts for this device
			IEC3CLR = _IEC3_U1RXIE_MASK;
			IEC3CLR = _IEC3_U1EIE_MASK;
            break;

        case 1:
			IEC4CLR = _IEC4_U2TXIE_MASK;
			IEC4CLR = _IEC4_U2RXIE_MASK;
			IEC4CLR = _IEC4_U2EIE_MASK;
            break;

        case 2:
			IEC4CLR = _IEC4_U3TXIE_MASK;
			IEC4CLR = _IEC4_U3RXIE_MASK;
			IEC4CLR = _IEC4_U3EIE_MASK;
            break;

        case 3:
			IEC5CLR = _IEC5_U4TXIE_MASK;
			IEC5CLR = _IEC5_U4RXIE_MASK;
			IEC5CLR = _IEC5_U4EIE_MASK;
            break;

        default: // do nothing if its not a valid port
            break;
    }
}


#if defined UART1

/* ********************* UART 1 Interrupt handlers ********************* */

void vU1RXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Are any Rx interrupts pending? */
	if( IFS3bits.U1RXIF == 1 )
	{
         /* Receive all data from RX FIFO that avilable*/
         while( U1STAbits.URXDA == 1)
		{
			/* Retrieve the received character and place it in the queue of
			     received characters. */
			cChar = U1RXREG;
			xQueueSendFromISR( xPorts[0].xRxedChars, &cChar, &xHigherPriorityTaskWoken );
			 /* clear any RX errors */		// <---- need to check interaction with ERR interrupt
			//if (U1STAbits.OERR) U1STAbits.OERR = 0;
			//if (U1STAbits.FERR) U1STAbits.FERR = 0;
			//if (U1STAbits.PERR) U1STAbits.PERR = 0;
		}
	}
	/* Clear the RX interrupt pending Flag */
	IFS3CLR = _IFS3_U1RXIF_MASK;
    
	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


void vU1TXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Check if UART TX interrupts pending */
    if (  IFS3bits.U1TXIF == 1)
	{
		while( !( U1STAbits.UTXBF ) ) /* Continue until UART TX fifo is full */
			{
				if( xQueueReceiveFromISR( xPorts[0].xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
					{
						/* Send the next character queued for Tx. */
						U1TXREG = cChar;
					}
					else
					{
						/* Queue empty, nothing to send. */
						/* Turn off UART TX interrupt */
						IEC3CLR = _IEC3_U1TXIE_MASK;
						xPorts[0].xTxHasEnded = pdTRUE;
						break;
					}
			}
	}
	IFS3CLR = _IFS3_U1TXIF_MASK;	// clear Tx interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

void vU1FAULTInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

  /* Any Rx errors interrupts pending? */
	if( IFS3bits.U1EIF == 1)
        {
            /* We need to clear any Overrun errors if any, or it will stop UART */
			if (U1STAbits.OERR) 
			{
				U1STAbits.OERR = 0;
				xPorts[0].ucOverrunErr++;
			}

			if (U1STAbits.FERR) 
			{
				U1STAbits.FERR = 0;
				xPorts[0].ucFramingErr++;
			}
			if (U1STAbits.PERR) 
			{
				U1STAbits.PERR = 0;
				xPorts[0].ucPartyErr++;
			}
        }


	IFS3CLR = _IFS3_U1EIF_MASK;	// clear Error interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


#endif

#if defined UART2

/* ********************* UART 2 Interrupt handlers ********************* */

void vU2RXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Are any Rx interrupts pending? */
	if( IFS4bits.U2RXIF == 1 )
	{
         /* Receive all data from RX FIFO that avilable*/
         while( U2STAbits.URXDA == 1)
		{
			/* Retrieve the received character and place it in the queue of
			  received characters. */
			cChar = U2RXREG;
			xQueueSendFromISR( xPorts[1].xRxedChars, &cChar, &xHigherPriorityTaskWoken );
			 /* clear any RX errors */
			//if (U2STAbits.OERR) U2STAbits.OERR = 0;
			//if (U2STAbits.FERR) U2STAbits.FERR = 0;
			//if (U2STAbits.PERR) U2STAbits.PERR = 0;
		}
	}
	/* Clear the RX interrupt pending Flag */
	IFS4CLR = _IFS4_U2RXIF_MASK;
    
	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


void vU2TXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Check if UART TX interrupts pending */
    if (  IFS4bits.U2TXIF == 1)
	{
		while( !( U2STAbits.UTXBF ) ) /* Continue until UART TX fifo is full */
			{
				if( xQueueReceiveFromISR( xPorts[1].xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
					{
						/* Send the next character queued for Tx. */
						U2TXREG = cChar;
					}
					else
					{
						/* Queue empty, nothing to send. */
						/* Turn off UART TX interrupt */
						IEC4CLR = _IEC4_U2TXIE_MASK;
						xPorts[1].xTxHasEnded = pdTRUE;
						break;
					}
			}
	}
	IFS4CLR = _IFS4_U2TXIF_MASK;	// clear Tx interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

void vU2FAULTInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Any Rx errors interrupts pending? */
	if( IFS4bits.U2EIF == 1)
        {
            /* We need to clear any Overrun errors if any, or it will stop UART */
			if (U2STAbits.OERR) 
			{
				U2STAbits.OERR = 0;
				xPorts[1].ucOverrunErr++;
			}

			if (U2STAbits.FERR) 
			{
				U2STAbits.FERR = 0;
				xPorts[1].ucFramingErr++;
			}
			if (U2STAbits.PERR) 
			{
				U2STAbits.PERR = 0;
				xPorts[1].ucPartyErr++;
			}
        }
	IFS4CLR = _IFS4_U2EIF_MASK;	// clear Error interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


#endif

/* ********************* UART 3 Interrupt handlers ********************* */
#if defined UART3

void vU3RXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Are any Rx interrupts pending? */
	if( IFS4bits.U3RXIF == 1 )
	{
         /* Receive all data from RX FIFO that avilable*/
         while( U3STAbits.URXDA == 1)
		{
			/* Retrieve the received character and place it in the queue of
			  received characters. */
			cChar = U3RXREG;
			xQueueSendFromISR( xPorts[2].xRxedChars, &cChar, &xHigherPriorityTaskWoken );
			 /* clear any RX errors */
			//if (U3STAbits.OERR) U3STAbits.OERR = 0;
			//if (U3STAbits.FERR) U3STAbits.FERR = 0;
			//if (U3STAbits.PERR) U3STAbits.PERR = 0;
		}
	}
	/* Clear the RX interrupt pending Flag */
	IFS4CLR = _IFS4_U3RXIF_MASK;
    
	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


void vU3TXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Check if UART TX interrupts pending */
    if (  IFS4bits.U3TXIF == 1)
	{
		while( !( U3STAbits.UTXBF ) ) /* Continue until UART TX fifo is full */
			{
				if( xQueueReceiveFromISR( xPorts[2].xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
					{
						/* Send the next character queued for Tx. */
						U3TXREG = cChar;
					}
					else
					{
						/* Queue empty, nothing to send. */
						/* Turn off UART TX interrupt */
						IEC4CLR = _IEC4_U3TXIE_MASK;
						xPorts[2].xTxHasEnded = pdTRUE;
						break;
					}
			}
	}
	IFS4CLR = _IFS4_U3TXIF_MASK;	// clear Tx interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

void vU3FAULTInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

  /* Any Rx errors interrupts pending? */
	if( IFS4bits.U3EIF == 1)
        {
            /* We need to clear any Overrun errors if any, or it will stop UART */
			if (U3STAbits.OERR) 
			{
				U3STAbits.OERR = 0;
				xPorts[2].ucOverrunErr++;
			}

			if (U1STAbits.FERR) 
			{
				U3STAbits.FERR = 0;
				xPorts[2].ucFramingErr++;
			}
			if (U3STAbits.PERR) 
			{
				U3STAbits.PERR = 0;
				xPorts[2].ucPartyErr++;
			}
        }
	IFS4CLR = _IFS4_U3EIF_MASK;	// clear Error interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

#endif

/* ********************* UART 4 Interrupt handlers ********************* */
#if defined UART4

void vU4RXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Are any Rx interrupts pending? */
	if( IFS5bits.U4RXIF == 1 )
	{
         /* Receive all data from RX FIFO that avilable*/
         while( U4STAbits.URXDA == 1)
		{
			/* Retrieve the received character and place it in the queue of
			     received characters. */
			cChar = U4RXREG;
			xQueueSendFromISR( xPorts[3].xRxedChars, &cChar, &xHigherPriorityTaskWoken );
			 /* clear any RX errors */		// <---- need to check interaction with ERR interrupt
			//if (U4STAbits.OERR) U4STAbits.OERR = 0;
			//if (U4STAbits.FERR) U4STAbits.FERR = 0;
			//if (U4STAbits.PERR) U4STAbits.PERR = 0;
		}
	}
	/* Clear the RX interrupt pending Flag */
	IFS5CLR = _IFS5_U4RXIF_MASK;

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


void vU4TXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Check if UART TX interrupts pending */
    if (  IFS5bits.U4TXIF == 1)
	{
		while( !( U4STAbits.UTXBF ) ) /* Continue until UART TX fifo is full */
			{
				if( xQueueReceiveFromISR( xPorts[3].xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
					{
						/* Send the next character queued for Tx. */
						U4TXREG = cChar;
					}
					else
					{
						/* Queue empty, nothing to send. */
						/* Turn off UART TX interrupt */
						IEC5CLR = _IEC5_U4TXIE_MASK;
						xPorts[3].xTxHasEnded = pdTRUE;
						break;
					}
			}
	}
	IFS5CLR = _IFS5_U4TXIF_MASK;	// clear Tx interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

void vU4FAULTInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

  /* Any Rx errors interrupts pending? */
	if( IFS5bits.U4EIF == 1)
        {
            /* We need to clear any Overrun errors if any, or it will stop UART */
			if (U4STAbits.OERR)
			{
				U4STAbits.OERR = 0;
				xPorts[0].ucOverrunErr++;
			}

			if (U4STAbits.FERR)
			{
				U4STAbits.FERR = 0;
				xPorts[0].ucFramingErr++;
			}
			if (U4STAbits.PERR)
			{
				U4STAbits.PERR = 0;
				xPorts[0].ucPartyErr++;
			}
        }


	IFS5CLR = _IFS5_U4EIF_MASK;	// clear Error interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


#endif


/* **************** NON RTOS Safe Serial I/O ****************** */

/*  These routines are NOT RTOS safe... use with care! 
 *	 required that the serial port be opened prior to any calls */
 
#if defined UART1

void uart1_puts(char *str) {
    while (*str != '\0') {
        while (U1STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U1TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart1_put(char *str, int len) {
    while (len-- > 0) {
        while (U1STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U1TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart1_putc(char c) {
    while (U1STAbits.UTXBF)
        ;
    U1TXREG = c;
}
/*-----------------------------------------------------------*/
char uart1_getc(void) {
    while (U1STAbits.URXDA == 0)
        ;
	return(U1RXREG);
}
/* ********************************************************* */

#endif

#if defined UART2
/*-----------------------------------------------------------*/
void uart2_puts(char *str) {
    while (*str != '\0') {
        while (U2STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U2TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart2_put(char *str, int len) {
    while (len-- > 0) {
        while (U2STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U2TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart2_putc(char c) {
    while (U2STAbits.UTXBF)
        ;
    U2TXREG = c;
}
/*-----------------------------------------------------------*/
char uart2_getc(void) {
    while (U2STAbits.URXDA == 0)
        ;
	return(U2RXREG);
}
/* ********************************************************* */
#endif

#if defined UART3

/*-----------------------------------------------------------*/
void uart3_puts(char *str) {
    while (*str != '\0') {
        while (U3STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U3TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart3_put(char *str, int len) {
    while (len-- > 0) {
        while (U3STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U3TXREG = *str++; /* transfer a data byte to TX reg */
    }
}
/*-----------------------------------------------------------*/
void uart3_putc(char c) {
    while (U3STAbits.UTXBF)
        ;
    U3TXREG = c;
}
/*-----------------------------------------------------------*/
char uart3_getc(void) {
    while (U3STAbits.URXDA == 0)
        ;
	return(U3RXREG);
}
/* ********************************************************* */
#endif

