
/*
*	TITLE		Queued serial drivers for FreeRTOS using PIC32MZ
*
*   File        serial_RTOS.c
*					
*	VERSION:	0.3b (Beta)
*
*	DATE:		1-Dec-2013
*
*	AUTHOR:		Tom Lafleur  
*				lafleur@lafleur.us
*
*	COMMENTS:	You are free to use this source code for any non-commercial
*				use. Please do not make copies of this source code, modified
*				or un-modified, publicly available on the internet or
*				elsewhere without permission. Thanks.
*
*			Copyright 2005-2014  All rights are reserved.
*
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	12-Jun-2009  0.1  TRL - Original code for N2K60 under FreeRTOS
*	01-Oct-2011  0.2a TRL - Added serial port 2, write function, error counters
*	15-Mar-2012  0.2b TRL - Clean up issues with .h files
*	03-Dec-2013	 0.3b TRL - Port to PIC32MZ processor
*
*
*	Notes:	1)  Tested with FreeRTOS ver 7.6.0 MEB II
*			2)  Support for two serial ports, can easily be expanded for more ports
*			3)  Added Rx error counters to the data structures
*			4)  If using processors with Peripheral pin select, you must define
*               pin's prior to using this driver <-----------------------------------
*			5)  Based on FreeRTOS demo code and my 2007 robotics serial driver  
*
*	ToDo:	1)  Check if code is atomic and thread safe
*			2)  Make Rx error counters avilable global
*			3)  Make Tx use DMA if available, if so we will use ring buffers
*               and give a full buffer to DMA system. Queue's are too slow.
*			4)
*
* 	My Notes:
*		
*		
*		
*/

#ifndef SERIAL_RTOS_H
#define SERIAL_RTOS_H

#include <xc.h>
#include "queue.h"
#include "serial.h"


/*---------------------------------------------------------------*/

typedef enum
{ 
    serCOM1,    // Only four ports supported at this time
	serCOM2,
	serCOM3,
	serCOM4
//	serCOM5,
//	serCOM6
} eCOMPort;

typedef enum 
{ 
	serNO_PARITY, 
	serEVEN_PARITY, 
	serODD_PARITY 
} eParity;

typedef enum 
{ 
	serSTOP_1, 
	serSTOP_2 
} eStopBits;

typedef enum 
{ 
	serBITS_8, 
	serBITS_9 
} eDataBits;

typedef struct xCOM_PORT {
    eCOMPort sPort; 						// Comm port address 0, 1 ect...
    xQueueHandle xRxedChars;                // Read buffer
    xQueueHandle xCharsForTx;               // Write buffer
    volatile portBASE_TYPE xTxHasEnded;    	// Tx empty flag
    volatile unsigned char ucPartyErr;      // Party error counter
    volatile unsigned char ucFramingErr;    // Framing error counter
    volatile unsigned char ucOverrunErr;    // Overrun error counter
} volatile xComPort;

typedef xComPort * xComPortHandle;

/*-----------------------------------------------------------*/

/*     Public Prototypes     */

/* Serial port intinalization */
xComPortHandle xSerialPortInit( eCOMPort ePort, unsigned long eWantedBaud, eParity eWantedParity, eDataBits eWantedDataBits, eStopBits eWantedStopBits, unsigned portBASE_TYPE uxTxBufferLength, unsigned portBASE_TYPE uxRxBufferLength );
/* Serial port GetChar, xBlockTime not used*/
signed portBASE_TYPE xSerialGetChar( xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime );
/* Serial Get String untill NULL, CR or usStringLength */
void xSerialGetString(xComPortHandle pxPort, signed char * pcString, unsigned short usStringLength, portTickType xBlockTime);
/* Serial Get String untill NULL, CR or usStringLength for FreeRTOS CLI*/
void xSerialGetStringS(xComPortHandle pxPort, signed char * pcString, unsigned short usStringLength, portTickType xBlockTime);
/* Serial Peek at next char in queue */
signed portBASE_TYPE xSerialPeekChar(xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime);
/* Serial port PutChar */
signed portBASE_TYPE xSerialPutChar( xComPortHandle pxPort, signed char cOutChar, portTickType xBlockTime );
/* Serial port PutString */
void vSerialPutString( xComPortHandle pxPort, const signed char * const pcString, unsigned short usStringLength );
/* Serial port PutString of length usStringLength */
void vSerialPutStringN(xComPortHandle pxPort, const signed char * const pcString, unsigned short usStringLength);
/* Serial Port Close */
void vSerialClose( xComPortHandle xPort );
/* Serial Port write, allow use of standard C I/O functions */
int write(int handle, void *buffer, unsigned int len);


/* ***** These functions are NOT RTOS safe, use with care!! ***** */
void uart1_puts(char *str);
void uart1_putc(char c);
void uart1_put(char *str, int len);
char uart1_getc(void);

void uart2_puts(char *str);
void uart2_putc(char c);
void uart2_put(char *str, int len);
char uart2_getc(void);

void uart3_puts(char *str);
void uart3_putc(char c);
void uart3_put(char *str, int len);
char uart3_getc(void);

void uart4_puts(char *str);
void uart4_putc(char c);
void uart4_put(char *str, int len);
char uart4_getc(void);

#endif

